# Copyright (c) 2006-2007 The Regents of The University of Michigan
# Copyright (c) 2009 Advanced Micro Devices, Inc.
# Copyright (c) 2011 Mark D. Hill and David A. Wood
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met: redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer;
# redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution;
# neither the name of the copyright holders nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# Authors: Brad Beckmann

import m5
from m5.objects import *
from m5.defines import buildEnv
from m5.util import panic
import math

class L1Cache(RubyCache):
	latency = 3

def create_system(options, system, cpu_sequencers, all_cntrls):

	if not buildEnv['GPGPU_SIM']:
		panic("This script requires the GPGPU-sim to be built.")

	#
	# The ruby network creation expects the list of nodes in the system to be
	# consistent with the NetDest list.  Therefore the l1 controller nodes must be
	# listed before the directory nodes and directory nodes before dma nodes, etc.
	#
	l1_cntrl_nodes = []

	print "making", options.num_sc, "sc's"

	#cache for stream processor
	#
	for i in xrange(options.num_sc):
		print "doing", i, "len:", len(cpu_sequencers)
		#
		# First create the Ruby objects associated with this cpu
		#
		l1i_cache = L1Cache(size = options.l1i_size,
							assoc = options.l1i_assoc)
		l1d_cache = L1Cache(size = options.l1d_size,
							assoc = options.l1d_assoc)

		cpu_seq = RubySequencer(version = options.num_cpus+i,
								icache = l1i_cache,
								dcache = l1d_cache,
								physMemPort = system.physmem.port,
								physmem = system.physmem,
                                access_phys_mem = False)

		l1_cntrl = L1Cache_Controller(version = options.num_cpus+i,
									sequencer = cpu_seq,
									L1IcacheMemory = l1i_cache,
									L1DcacheMemory = l1d_cache,
									l2_select_num_bits = \
										math.log(options.num_l2caches,
												2))

		exec("system.l1_cntrlsp%d = l1_cntrl" % i)
		#
		# Add controllers and sequencers to the appropriate lists
		#
		cpu_sequencers.append(cpu_seq)
		l1_cntrl_nodes.append(l1_cntrl)

	all_cntrls += l1_cntrl_nodes

	return
