# Copyright (c) 2006-2007 The Regents of The University of Michigan
# Copyright (c) 2009 Advanced Micro Devices, Inc.
# Copyright (c) 2011 Mark D. Hill and David A. Wood
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met: redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer;
# redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution;
# neither the name of the copyright holders nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# Authors: Brad Beckmann

import m5
from m5.objects import *
from m5.defines import buildEnv

class Cache(RubyCache):
    latency = 14

def create_system(options, system, cpu_sequencers, all_cntrls, ruby_system):

    if not buildEnv['GPGPU_SIM']:
        panic("This script requires the GPGPU-sim to be built.")

    #
    # The ruby network creation expects the list of nodes in the system to be
    # consistent with the NetDest list.  Therefore the l1 controller nodes must be
    # listed before the directory nodes and directory nodes before dma nodes, etc.
    #
    l1_cntrl_nodes = []


    #cache for stream processor
    #
    for i in xrange(options.num_sc):
       # First create the Ruby objects associated with this cpu
       # Only one cache exists for this protocol, so by default use the L1D
       # config parameters.
       #
       cache = Cache(size = options.sc_l1_size,
           assoc = options.sc_l1_assoc, replacement_policy = "LRU")

       #
       # Only one unified L1 cache exists.  Can cache instructions and data.
       #
       cpu_seq = RubySequencer(version = options.num_cpus+i,
                               icache = cache,
                               dcache = cache,
                               physMemPort = system.physmem.port,
                               physmem = system.physmem,
                               access_phys_mem = False,
                               ruby_system = ruby_system)

       l1_cntrl = L1Cache_Controller(version = options.num_cpus+i,
                                       cntrl_id = len(all_cntrls)+i,
                                       sequencer = cpu_seq,
                                       cacheMemory = cache,
                                       ruby_system = ruby_system)

       #exec("system.l1_cntrl_sp = l1_cntrl")
       if (i<10):
           exec("system.l1_cntrl_sp0%d = l1_cntrl" % i)
       else:
           exec("system.l1_cntrl_sp%d = l1_cntrl" % i)
       #
       # Add controllers and sequencers to the appropriate lists
       #
       cpu_sequencers.append(cpu_seq)
       l1_cntrl_nodes.append(l1_cntrl)

	######################################################################################
    #copy engine cache (make as small as possible, ideally 0)
    cache = Cache(size = "4kB", assoc = 2)

    #
    # Only one unified L1 cache exists.  Can cache instructions and data.
    #
    cpu_seq = RubySequencer(version = options.num_cpus+options.num_sc,
                               icache = cache,
                               dcache = cache,
                               physMemPort = system.physmem.port,
                               physmem = system.physmem,
                               access_phys_mem = True,
                               ruby_system = ruby_system)

    l1_cntrl = L1Cache_Controller(version = options.num_cpus+options.num_sc,
                                    cntrl_id = len(all_cntrls)+options.num_sc,
                                    sequencer = cpu_seq,
                                    cacheMemory = cache,
                                    ruby_system = ruby_system)

    system.l1_cntrl_ce = l1_cntrl

    cpu_sequencers.append(cpu_seq)
    l1_cntrl_nodes.append(l1_cntrl)

    all_cntrls += l1_cntrl_nodes

    return
