// Copyright (c) 2009-2011, The University of British Columbia
// Copyright (c) 2011 Mark D. Hill and David A. Wood
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice, this
// list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice, this
// list of conditions and the following disclaimer in the documentation and/or
// other materials provided with the distribution.
// Neither the name of The University of British Columbia nor the names of its
// contributors may be used to endorse or promote products derived from this
// software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
// ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
// WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
// DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
// FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
// DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
// SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
// CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
// OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
// OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#include "memory_gem5.h"
#include "../debug.h"

// NOTE: mem_addr_t is an unsigned int

void memory_gem5::write( mem_addr_t addr, size_t length, const void *data, ptx_thread_info *thd, const ptx_instruction *pI )
{
    abstractGPU->gem5_spa->writeFunctional(addr, length, (uint8_t*)data);

    //copied from memory_space_impl<BSIZE>::write
    if( !m_watchpoints.empty() ) {
        std::map<unsigned,mem_addr_t>::iterator i;
        for( i=m_watchpoints.begin(); i!=m_watchpoints.end(); i++ ) {
            mem_addr_t wa = i->second;
            if( ((addr<=wa) && ((addr+length)>wa)) || ((addr>wa) && (addr < (wa+4))) )
                hit_watchpoint(i->first,thd,pI);
        }
    }
    std::map<unsigned,mem_addr_t> m_watchpoints;
}


void memory_gem5::read( mem_addr_t addr, size_t length, void *data ) const
{
    abstractGPU->gem5_spa->readFunctional(addr, length, (uint8_t*)data);
}

void memory_gem5::set_watch( addr_t addr, unsigned watchpoint )
{
   m_watchpoints[watchpoint]=addr;
}

