#ifndef _KERNEL_HOST_H_
#define _KERNEL_HOST_H_


static const double one=1.0, two=2.0, tiny = 1.0e-300;

typedef union {
	double value;
	struct  {
		int msw;
		int lsw;
	} parts;
} ieee_double_shape_type;

#define GET_HIGH_WORD(d,v)                                      \
do {                                                            \
  ieee_double_shape_type sh_u;                                  \
  sh_u.value = (d);                                         \
  (v) = sh_u.parts.msw;                                             \
} while (0)

#define fabs(x) (((x)<0)?-1.0*(x):(x))

double tanh(double x) __THROW
{
	double t,z;
	int32_t jx,ix;

	/* High word of |x|. */
	GET_HIGH_WORD(jx,x);
	ix = jx&0x7fffffff;

	/* x is INF or NaN */
	if(ix>=0x7ff00000) {
		if (jx>=0) return one/x+one;    /* tanh(+-inf)=+-1 */
		else       return one/x-one;    /* tanh(NaN) = NaN */
	}

	/* |x| < 22 */
	if (ix < 0x40360000) {          /* |x|<22 */
		if (ix<0x3c800000)          /* |x|<2**-55 */
			return x*(one+x);       /* tanh(small) = small */
		if (ix>=0x3ff00000) {       /* |x|>=1  */
			t = expm1(two*fabs(x));
			z = one - two/(t+two);
		} else {
			t = expm1(-two*fabs(x));
			z= -t/(t+two);
		}
		/* |x| > 22, return +-1 */
	} else {
		z = one - tiny;             /* raised inexact flag */
	}
	return (jx>=0)? z: -z;
}

void executeThirdLayerHost(float *Layer3_Neurons_GPU, float *Layer3_Weights_GPU,float *Layer4_Neurons_GPU, int NUM)
{
	for (int blockID=0; blockID<100; blockID++) {
		for (int blockIdy=0; blockIdy<NUM; blockIdy++) {
			int weightBegin=blockID*1251;

			float result=0;

			result+=Layer3_Weights_GPU[weightBegin];

			++weightBegin;

			for (int i=0; i<1250; ++i )
			{
				result+=Layer3_Neurons_GPU[i+(1250*blockIdy)]*Layer3_Weights_GPU[weightBegin+i];
			}

			result=(1.7159*tanh(0.66666667*result));

			Layer4_Neurons_GPU[blockID+(100*blockIdy)]=result;
		}
	}

}

void executeFourthLayerHost(float *Layer4_Neurons_GPU,float *Layer4_Weights_GPU,float *Layer5_Neurons_GPU, int NUM)
{
	for (int blockID=0; blockID<10; blockID++) {
		for (int blockIdy=0; blockIdy<NUM; blockIdy++) {
			int weightBegin=blockID*101;

			float result=0;

			result+=Layer4_Weights_GPU[weightBegin];

			++weightBegin;

			for (int i=0; i<100; ++i )
			{
				result+=Layer4_Neurons_GPU[i+(100*blockIdy)]*Layer4_Weights_GPU[weightBegin+i];
			}

			result=(1.7159*tanh(0.66666667*result));

			Layer5_Neurons_GPU[blockID+(10*blockIdy)]=result;
		}
	}
}

#endif
