// This file created from cuda_runtime_api.h distributed with CUDA 1.1
// Changes Copyright 2009,  Tor M. Aamodt, Ali Bakhoda and George L. Yuan
// University of British Columbia

/*
 * cuda_runtime_api.cc
 *
 * Copyright © 2009 by Tor M. Aamodt, Wilson W. L. Fung, Ali Bakhoda,
 * George L. Yuan and the University of British Columbia, Vancouver,
 * BC V6T 1Z4, All Rights Reserved.
 *
 * THIS IS A LEGAL DOCUMENT BY DOWNLOADING GPGPU-SIM, YOU ARE AGREEING TO THESE
 * TERMS AND CONDITIONS.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNERS OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * NOTE: The files libcuda/cuda_runtime_api.c and src/cuda-sim/cuda-math.h
 * are derived from the CUDA Toolset available from http://www.nvidia.com/cuda
 * (property of NVIDIA).  The files benchmarks/BlackScholes/ and
 * benchmarks/template/ are derived from the CUDA SDK available from
 * http://www.nvidia.com/cuda (also property of NVIDIA).  The files from
 * src/intersim/ are derived from Booksim (a simulator provided with the
 * textbook "Principles and Practices of Interconnection Networks" available
 * from http://cva.stanford.edu/books/ppin/). As such, those files are bound by
 * the corresponding legal terms and conditions set forth separately (original
 * copyright notices are left in files from these sources and where we have
 * modified a file our copyright notice appears before the original copyright
 * notice).
 *
 * Using this version of GPGPU-Sim requires a complete installation of CUDA
 * which is distributed seperately by NVIDIA under separate terms and
 * conditions.  To use this version of GPGPU-Sim with OpenCL requires a
 * recent version of NVIDIA's drivers which support OpenCL.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimer in the documentation
 * and/or other materials provided with the distribution.
 *
 * 3. Neither the name of the University of British Columbia nor the names of
 * its contributors may be used to endorse or promote products derived from
 * this software without specific prior written permission.
 *
 * 4. This version of GPGPU-SIM is distributed freely for non-commercial use only.
 *
 * 5. No nonprofit user may place any restrictions on the use of this software,
 * including as modified by the user, by any other authorized user.
 *
 * 6. GPGPU-SIM was developed primarily by Tor M. Aamodt, Wilson W. L. Fung,
 * Ali Bakhoda, George L. Yuan, at the University of British Columbia,
 * Vancouver, BC V6T 1Z4
 */

/*
 * Copyright 1993-2007 NVIDIA Corporation.  All rights reserved.
 *
 * NOTICE TO USER:
 *
 * This source code is subject to NVIDIA ownership rights under U.S. and
 * international Copyright laws.  Users and possessors of this source code
 * are hereby granted a nonexclusive, royalty-free license to use this code
 * in individual and commercial software.
 *
 * NVIDIA MAKES NO REPRESENTATION ABOUT THE SUITABILITY OF THIS SOURCE
 * CODE FOR ANY PURPOSE.  IT IS PROVIDED "AS IS" WITHOUT EXPRESS OR
 * IMPLIED WARRANTY OF ANY KIND.  NVIDIA DISCLAIMS ALL WARRANTIES WITH
 * REGARD TO THIS SOURCE CODE, INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY, NONINFRINGEMENT, AND FITNESS FOR A PARTICULAR PURPOSE.
 * IN NO EVENT SHALL NVIDIA BE LIABLE FOR ANY SPECIAL, INDIRECT, INCIDENTAL,
 * OR CONSEQUENTIAL DAMAGES, OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS
 * OF USE, DATA OR PROFITS,  WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE
 * OR OTHER TORTIOUS ACTION,  ARISING OUT OF OR IN CONNECTION WITH THE USE
 * OR PERFORMANCE OF THIS SOURCE CODE.
 *
 * U.S. Government End Users.   This source code is a "commercial item" as
 * that term is defined at  48 C.F.R. 2.101 (OCT 1995), consisting  of
 * "commercial computer  software"  and "commercial computer software
 * documentation" as such terms are  used in 48 C.F.R. 12.212 (SEPT 1995)
 * and is provided to the U.S. Government only as a commercial end item.
 * Consistent with 48 C.F.R.12.212 and 48 C.F.R. 227.7202-1 through
 * 227.7202-4 (JUNE 1995), all U.S. Government End Users acquire the
 * source code with only those rights set forth herein.
 *
 * Any use of this source code in individual and commercial software must
 * include, in the user documentation and internal comments to the code,
 * the above Disclaimer and U.S. Government End Users Notice.
 */
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <assert.h>
#include <time.h>
#include <stdarg.h>

#include <unistd.h>
#include <stdint.h>

#ifdef OPENGL_SUPPORT
#define GL_GLEXT_PROTOTYPES
#include <GL/gl.h>
#endif

#define __CUDA_RUNTIME_API_H__

/*******************************************************************************
*                                                                              *
*                                                                              *
*                                                                              *
*******************************************************************************/


#include "host_defines.h"
#include "builtin_types.h"
#include "__cudaFatFormat.h"

/*DEVICE_BUILTIN*/
struct cudaArray
{
    void *devPtr;
    int devPtr32;
    struct cudaChannelFormatDesc desc;
    int width;
    int height;
    int size; //in bytes
    unsigned dimensions;
};


#if !defined(__dv)
#if defined(__cplusplus)
#define __dv(v) \
        = v
#else /* __cplusplus */
#define __dv(v)
#endif /* __cplusplus */
#endif /* !__dv */

/*******************************************************************************
*                                                                              *
*                                                                              *
*                                                                              *
*******************************************************************************/

struct gpgpu_ptx_sim_arg {
   const void *m_start;
   size_t m_nbytes;
   size_t m_offset;
   struct gpgpu_ptx_sim_arg *m_next;
};

# if defined __cplusplus ? __GNUC_PREREQ (2, 6) : __GNUC_PREREQ (2, 4)
#   define __my_func__    __PRETTY_FUNCTION__
# else
#  if defined __STDC_VERSION__ && __STDC_VERSION__ >= 199901L
#   define __my_func__    __func__
#  else
#   define __my_func__    ((__const char *) 0)
#  endif
# endif



/*
 *  bytes 			= array to store parameters in byte form
 *  bytes_off 		= offset into bytes array
 *  lengths 		= array or integers, each int is the size of an argument
 *  lengths_off 	= offset into arg_sizes
 *  arg 				= pointer to parameter
 *  arg_size 		= size of argument
 */
void pack(char *bytes, int &bytes_off, int *lengths, int &lengths_off, char *arg, int arg_size)
{
	//arg = arg+arg_size-1; //push last bytes of arg onto stack first
	for(int i=0; i<arg_size; i++)
	{
		//args.push(*arg);
		bytes[bytes_off + i] = *arg;
		arg++;
	}
	*(lengths+lengths_off) = arg_size;

	bytes_off += arg_size;
	lengths_off += 1;
}

char *unpack(char *bytes, int &bytes_off, int *lengths, int &lengths_off)
{
	int arg_size = *(lengths+lengths_off);
	char *arg = new char[arg_size];
	for(int i=0; i<arg_size; i++)
	{
		arg[i] = bytes[i+bytes_off];
	}

	bytes_off += arg_size;
	lengths_off += 1;

	return arg;
}


extern "C" {

/*******************************************************************************
*                                                                              *
*                                                                              *
*                                                                              *
*******************************************************************************/

__host__ cudaError_t CUDARTAPI cudaMalloc(void **devPtr, size_t size)
{
    cudaError_t ret = (cudaError_t)syscall(273, 0, devPtr, size);
    if(ret == cudaErrorApiFailureBase) {
        posix_memalign(devPtr, 128, size);
        ret = cudaSuccess;
    }

    return ret;
}

__host__ cudaError_t CUDARTAPI cudaMallocHost(void **ptr, size_t size){
	return (cudaError_t)syscall(273, 1, ptr, size);
}

 __host__ cudaError_t CUDARTAPI cudaMallocPitch(void **devPtr, size_t *pitch, size_t width, size_t height)
{
	return (cudaError_t)syscall(273, 2, devPtr, pitch, width, height);
}

__host__ cudaError_t CUDARTAPI cudaMallocArray(struct cudaArray **array, const struct cudaChannelFormatDesc *desc, size_t width, size_t height __dv(1))
{
	return (cudaError_t)syscall(273, 3, array, desc, width, height);
}

__host__ cudaError_t CUDARTAPI cudaFree(void *devPtr)
{
	return (cudaError_t)syscall(273, 4, devPtr);
}
 __host__ cudaError_t CUDARTAPI cudaFreeHost(void *ptr)
{
	return (cudaError_t)syscall(273, 5, ptr);
}

 __host__ cudaError_t CUDARTAPI cudaFreeArray(struct cudaArray *array)
{
	return (cudaError_t)syscall(273, 6, array);
};


/*******************************************************************************
*                                                                              *
*                                                                              *
*                                                                              *
*******************************************************************************/

 __host__ cudaError_t CUDARTAPI cudaMemcpy(void *dst, const void *src, size_t count, enum cudaMemcpyKind kind)
{
	return (cudaError_t)syscall(273, 7, dst, src, count, kind);
	//replace with memcpy, so we can get the timing
	//memcpy( dst, src, count );
	//return (cudaError_t)0;
}

 __host__ cudaError_t CUDARTAPI cudaMemcpyToArray(struct cudaArray *dst, size_t wOffset, size_t hOffset, const void *src, size_t count, enum cudaMemcpyKind kind)
{
	return (cudaError_t)syscall(273, 8, dst, wOffset, hOffset, src, count, kind);
}


 __host__ cudaError_t CUDARTAPI cudaMemcpyFromArray(void *dst, const struct cudaArray *src, size_t wOffset, size_t hOffset, size_t count, enum cudaMemcpyKind kind)
{
	return (cudaError_t)syscall(273, 9, dst, src, wOffset, hOffset, count, kind);
}


 __host__ cudaError_t CUDARTAPI cudaMemcpyArrayToArray(struct cudaArray *dst, size_t wOffsetDst, size_t hOffsetDst, const struct cudaArray *src, size_t wOffsetSrc, size_t hOffsetSrc, size_t count, enum cudaMemcpyKind kind __dv(cudaMemcpyDeviceToDevice))
{
	return (cudaError_t)syscall(273, 10, dst, wOffsetDst, hOffsetDst, src, count, wOffsetSrc, hOffsetSrc, count, kind);
}


 __host__ cudaError_t CUDARTAPI cudaMemcpy2D(void *dst, size_t dpitch, const void *src, size_t spitch, size_t width, size_t height, enum cudaMemcpyKind kind)
{
	return (cudaError_t)syscall(273, 11, dst, dpitch, src, spitch, width, height, kind);
}


 __host__ cudaError_t CUDARTAPI cudaMemcpy2DToArray(struct cudaArray *dst, size_t wOffset, size_t hOffset, const void *src, size_t spitch, size_t width, size_t height, enum cudaMemcpyKind kind)
{
	return (cudaError_t)syscall(273, 12, dst, wOffset, hOffset, src, spitch, width, height, kind);
}


 __host__ cudaError_t CUDARTAPI cudaMemcpy2DFromArray(void *dst, size_t dpitch, const struct cudaArray *src, size_t wOffset, size_t hOffset, size_t width, size_t height, enum cudaMemcpyKind kind)
{
	return (cudaError_t)syscall(273, 13, dst, dpitch, src, wOffset, hOffset, width, height, kind);
}


 __host__ cudaError_t CUDARTAPI cudaMemcpy2DArrayToArray(struct cudaArray *dst, size_t wOffsetDst, size_t hOffsetDst, const struct cudaArray *src, size_t wOffsetSrc, size_t hOffsetSrc, size_t width, size_t height, enum cudaMemcpyKind kind __dv(cudaMemcpyDeviceToDevice))
{
	return (cudaError_t)syscall(273, 14, wOffsetDst, hOffsetDst, src, wOffsetSrc, hOffsetSrc, width, height, kind);
}


 __host__ cudaError_t CUDARTAPI cudaMemcpyToSymbol(const char *symbol, const void *src, size_t count, size_t offset __dv(0), enum cudaMemcpyKind kind __dv(cudaMemcpyHostToDevice))
{
	return (cudaError_t)syscall(273, 15, symbol, src, count, offset, kind);
}


 __host__ cudaError_t CUDARTAPI cudaMemcpyFromSymbol(void *dst, const char *symbol, size_t count, size_t offset __dv(0), enum cudaMemcpyKind kind __dv(cudaMemcpyDeviceToHost))
{
	return (cudaError_t)syscall(273, 16, dst, symbol, count, offset, kind);
}



/*******************************************************************************
*                                                                              *
*                                                                              *
*                                                                              *
*******************************************************************************/

 __host__ cudaError_t CUDARTAPI cudaMemcpyAsync(void *dst, const void *src, size_t count, enum cudaMemcpyKind kind, cudaStream_t stream)
{
	return (cudaError_t)syscall(273, 17, dst, src, count, kind, stream);
}


 __host__ cudaError_t CUDARTAPI cudaMemcpyToArrayAsync(struct cudaArray *dst, size_t wOffset, size_t hOffset, const void *src, size_t count, enum cudaMemcpyKind kind, cudaStream_t stream)
{
	return (cudaError_t)syscall(273, 18, dst, src, count, kind, stream);
}


 __host__ cudaError_t CUDARTAPI cudaMemcpyFromArrayAsync(void *dst, const struct cudaArray *src, size_t wOffset, size_t hOffset, size_t count, enum cudaMemcpyKind kind, cudaStream_t stream)
{
	return (cudaError_t)syscall(273, 19, dst, src, wOffset, hOffset, count, kind, stream);
}


 __host__ cudaError_t CUDARTAPI cudaMemcpy2DAsync(void *dst, size_t dpitch, const void *src, size_t spitch, size_t width, size_t height, enum cudaMemcpyKind kind, cudaStream_t stream)
{
	return (cudaError_t)syscall(273, 20, dst, dpitch, src, spitch, width, height, kind, stream);
}


 __host__ cudaError_t CUDARTAPI cudaMemcpy2DToArrayAsync(struct cudaArray *dst, size_t wOffset, size_t hOffset, const void *src, size_t spitch, size_t width, size_t height, enum cudaMemcpyKind kind, cudaStream_t stream)
{
	return (cudaError_t)syscall(273, 21, dst, wOffset, hOffset, src, spitch, width, height, kind, stream);
}


 __host__ cudaError_t CUDARTAPI cudaMemcpy2DFromArrayAsync(void *dst, size_t dpitch, const struct cudaArray *src, size_t wOffset, size_t hOffset, size_t width, size_t height, enum cudaMemcpyKind kind, cudaStream_t stream)
{
	return (cudaError_t)syscall(273, 22, dst, dpitch, src, wOffset, hOffset, width, height, kind, stream);
}



/*******************************************************************************
*                                                                              *
*                                                                              *
*                                                                              *
*******************************************************************************/

__host__ cudaError_t CUDARTAPI cudaMemset(void *mem, int c, size_t count)
{
	return (cudaError_t)syscall(273, 23, mem, c, count);
}

 __host__ cudaError_t CUDARTAPI cudaMemset2D(void *mem, size_t pitch, int c, size_t width, size_t height)
{
	return (cudaError_t)syscall(273, 24, mem, pitch, c, width, height);
}



/*******************************************************************************
*                                                                              *
*                                                                              *
*                                                                              *
*******************************************************************************/

 __host__ cudaError_t CUDARTAPI cudaGetSymbolAddress(void **devPtr, const char *symbol)
{
	return (cudaError_t)syscall(273, 25, devPtr, symbol);
}


 __host__ cudaError_t CUDARTAPI cudaGetSymbolSize(size_t *size, const char *symbol)
{
	return (cudaError_t)syscall(273, 26, size, symbol);
}



/*******************************************************************************
*                                                                              *
*                                                                              *
*                                                                              *
*******************************************************************************/
 __host__ cudaError_t CUDARTAPI cudaGetDeviceCount(int *count)
{
	return (cudaError_t)syscall(273, 27, count);
}

 __host__ cudaError_t CUDARTAPI cudaGetDeviceProperties(struct cudaDeviceProp *prop, int device)
{
	return (cudaError_t)syscall(273, 28, prop, device);
}

 __host__ cudaError_t CUDARTAPI cudaChooseDevice(int *device, const struct cudaDeviceProp *prop)
{
	return (cudaError_t)syscall(273, 29, device, prop);
}

 __host__ cudaError_t CUDARTAPI cudaSetDevice(int device)
{
	return (cudaError_t)syscall(273, 30, device);
}

 __host__ cudaError_t CUDARTAPI cudaGetDevice(int *device)
{
	return (cudaError_t)syscall(273, 31, device);
}

/*******************************************************************************
*                                                                              *
*                                                                              *
*                                                                              *
*******************************************************************************/

 __host__ cudaError_t CUDARTAPI cudaBindTexture(size_t *offset, const struct textureReference *texref, const void *devPtr, const struct cudaChannelFormatDesc *desc, size_t size __dv(UINT_MAX))
{
	return (cudaError_t)syscall(273, 32, offset, texref, devPtr, desc, size);
}


 __host__ cudaError_t CUDARTAPI cudaBindTextureToArray(const struct textureReference *texref, const struct cudaArray *array, const struct cudaChannelFormatDesc *desc)
{
	return (cudaError_t)syscall(273, 33, texref, array, desc);
}


 __host__ cudaError_t CUDARTAPI cudaUnbindTexture(const struct textureReference *texref)
{
	return (cudaError_t)syscall(273, 34, texref);
}


 __host__ cudaError_t CUDARTAPI cudaGetTextureAlignmentOffset(size_t *offset, const struct textureReference *texref)
{
	return (cudaError_t)syscall(273, 35, offset, texref);
}


 __host__ cudaError_t CUDARTAPI cudaGetTextureReference(const struct textureReference **texref, const char *symbol)
{
	return (cudaError_t)syscall(273, 36, texref, symbol);
}



/*******************************************************************************
*                                                                              *
*                                                                              *
*                                                                              *
*******************************************************************************/

 __host__ cudaError_t CUDARTAPI cudaGetChannelDesc(struct cudaChannelFormatDesc *desc, const struct cudaArray *array)
{
	return (cudaError_t)syscall(273, 37, desc, array);
}


 __host__ struct cudaChannelFormatDesc CUDARTAPI cudaCreateChannelDesc(int x, int y, int z, int w, enum cudaChannelFormatKind f)
{
	struct cudaChannelFormatDesc dummy;
	dummy.x = x;
	dummy.y = y;
	dummy.z = z;
	dummy.w = w;
	dummy.f = f;
	// NOTE had to change this
	//syscall(273, 38, x, y, z, w, f);
	return dummy;
}

/*******************************************************************************
*                                                                              *
*                                                                              *
*                                                                              *
*******************************************************************************/

 __host__ cudaError_t CUDARTAPI cudaGetLastError(void)
{
	return (cudaError_t)syscall(273, 39);
}

 __host__ const char* CUDARTAPI cudaGetErrorString(cudaError_t error)
{

	return (const char*)syscall(273, 40, error);
}



/*******************************************************************************
*                                                                              *
*                                                                              *
*                                                                              *
*******************************************************************************/

 __host__ cudaError_t CUDARTAPI cudaConfigureCall(dim3 gridDim, dim3 blockDim, size_t sharedMem, cudaStream_t stream)
{
	// NOTE had to change this!!
	return (cudaError_t)syscall(273, 41, &gridDim, &blockDim, sharedMem, stream);
}

__host__ cudaError_t CUDARTAPI cudaSetupArgument(const void *arg, size_t size, size_t offset){
	return (cudaError_t)syscall(273, 42, arg, size, offset);
}


__host__ cudaError_t CUDARTAPI cudaLaunch(const char *symbol )
{
	return (cudaError_t)syscall(273, 43, symbol);
}

/*******************************************************************************
*                                                                              *
*                                                                              *
*                                                                              *
*******************************************************************************/

__host__ cudaError_t CUDARTAPI cudaStreamCreate(cudaStream_t *stream)
{
	return (cudaError_t)syscall(273, 44, stream);
}

__host__ cudaError_t CUDARTAPI cudaStreamDestroy(cudaStream_t stream)
{
	return (cudaError_t)syscall(273, 45, stream);
}

__host__ cudaError_t CUDARTAPI cudaStreamSynchronize(cudaStream_t stream)
{
	return (cudaError_t)syscall(273, 46, stream);
}

__host__ cudaError_t CUDARTAPI cudaStreamQuery(cudaStream_t stream)
{
	return (cudaError_t)syscall(273, 47, stream);
}

/*******************************************************************************
*                                                                              *
*                                                                              *
*                                                                              *
*******************************************************************************/

__host__ cudaError_t CUDARTAPI cudaEventCreate(cudaEvent_t *event)
{
	return (cudaError_t)syscall(273, 48, event);
}

__host__ cudaError_t CUDARTAPI cudaEventRecord(cudaEvent_t event, cudaStream_t stream)
{
	return (cudaError_t)syscall(273, 49, event, stream);
}

__host__ cudaError_t CUDARTAPI cudaEventQuery(cudaEvent_t event)
{
	return (cudaError_t)syscall(273, 50, event);
}

__host__ cudaError_t CUDARTAPI cudaEventSynchronize(cudaEvent_t event)
{
	return (cudaError_t)syscall(273, 51, event);
}

__host__ cudaError_t CUDARTAPI cudaEventDestroy(cudaEvent_t event)
{
	return (cudaError_t)syscall(273, 52, event);
}


__host__ cudaError_t CUDARTAPI cudaEventElapsedTime(float *ms, cudaEvent_t start, cudaEvent_t end)
{
	return (cudaError_t)syscall(273, 53, ms, start, end);
}



/*******************************************************************************
*                                                                              *
*                                                                              *
*                                                                              *
*******************************************************************************/

__host__ cudaError_t CUDARTAPI cudaThreadExit(void)
{
	return (cudaError_t)syscall(273, 54);
}


__host__ cudaError_t CUDARTAPI cudaThreadSynchronize(void)
{
	return (cudaError_t)syscall(273, 55);
};

int CUDARTAPI __cudaSynchronizeThreads(void**, void*)
{
	return (cudaError_t)syscall(273, 56);
}


/*******************************************************************************
*                                                                              *
*                                                                              *
*                                                                              *
*******************************************************************************/

void** CUDARTAPI __cudaRegisterFatBinary( void *fatCubin )
{
	return (void**)syscall(273, 57, fatCubin, -1);
}

void** CUDARTAPI __cudaRegisterFatBinary2( void *fatCubin, size_t size )
{
	return (void**)syscall(273, 57, fatCubin, size);
}
void __cudaUnregisterFatBinary(void **fatCubinHandle)
{
	syscall(273, 58, fatCubinHandle);
}


void CUDARTAPI __cudaRegisterFunction(
                                     void   **fatCubinHandle,		//arg0
                                     const char    *hostFun,		//arg1
                                     char    *deviceFun,				//arg2
                                     const char    *deviceName,	//arg3
                                     int      thread_limit,			//arg4
                                     uint3   *tid,						//arg5
                                     uint3   *bid,						//arg6
                                     dim3    *bDim,					//arg7
                                     dim3    *gDim						//arg8
                                     )
{
	int num_args = 9;
	int arg0_size = sizeof(void **);
	int arg1_size = sizeof(const char *);
	int arg2_size = sizeof(char *);
	int arg3_size = sizeof(const char *);
	int arg4_size = sizeof(int);
	int arg5_size = sizeof(uint3 *);
	int arg6_size = sizeof(uint3 *);
	int arg7_size = sizeof(dim3 *);
	int arg8_size = sizeof(dim3 *);

	int total_bytes = arg0_size + arg1_size + arg2_size + arg3_size + arg4_size +
							arg5_size + arg6_size + arg7_size + arg8_size;

	//allocate space for packed parameters and their lengths
	char *bytes = new char[total_bytes];
	int *lengths = new int[num_args];

	//variables to keep track of packed parameters and their lengths
	int bytes_off = 0;
	int lengths_off = 0;

	//pack up the parameters
	pack(bytes, bytes_off, lengths, lengths_off, (char *)&fatCubinHandle, arg0_size);
	pack(bytes, bytes_off, lengths, lengths_off, (char *)&hostFun, arg1_size);
	pack(bytes, bytes_off, lengths, lengths_off, (char *)&deviceFun, arg2_size);
	pack(bytes, bytes_off, lengths, lengths_off, (char *)&deviceName, arg3_size);
	pack(bytes, bytes_off, lengths, lengths_off, (char *)&thread_limit, arg4_size);
	pack(bytes, bytes_off, lengths, lengths_off, (char *)&tid, arg5_size);
	pack(bytes, bytes_off, lengths, lengths_off, (char *)&bid, arg6_size);
	pack(bytes, bytes_off, lengths, lengths_off, (char *)&bDim, arg7_size);
	pack(bytes, bytes_off, lengths, lengths_off, (char *)&gDim, arg8_size);

	//make the new syscall
	//syscall(273, 59, fatCubinHandle, hostFun, deviceFun, deviceName, thread_limit, tid, bid, bDim, gDim);
	syscall(273, 59, num_args, lengths, bytes);

	delete bytes;
	delete lengths;
}

extern void __cudaRegisterVar(
                             void **fatCubinHandle,
                             char *hostVar, //pointer to...something
                             char *deviceAddress, //name of variable
                             const char *deviceName, //name of variable (same as above)
                             int ext,
                             int size,
                             int constant,
                             int global )
{
	int num_args = 8;
	int arg0_size = sizeof(void **);
	int arg1_size = sizeof(char *);
	int arg2_size = sizeof(char *);
	int arg3_size = sizeof(const char *);
	int arg4_size = sizeof(int);
	int arg5_size = sizeof(int);
	int arg6_size = sizeof(int);
	int arg7_size = sizeof(int);

	int total_bytes = arg0_size + arg1_size + arg2_size + arg3_size + arg4_size +
							arg5_size + arg6_size + arg7_size;

	//allocate space for packed parameters and their lengths
	char *bytes = new char[total_bytes];
	int *lengths = new int[num_args];

	//variables to keep track of packed parameters and their lengths
	int bytes_off = 0;
	int lengths_off = 0;

	//pack up the parameters
	pack(bytes, bytes_off, lengths, lengths_off, (char *)&fatCubinHandle, arg0_size);
	pack(bytes, bytes_off, lengths, lengths_off, (char *)&hostVar, arg1_size);
	pack(bytes, bytes_off, lengths, lengths_off, (char *)&deviceAddress, arg2_size);
	pack(bytes, bytes_off, lengths, lengths_off, (char *)&deviceName, arg3_size);
	pack(bytes, bytes_off, lengths, lengths_off, (char *)&ext, arg4_size);
	pack(bytes, bytes_off, lengths, lengths_off, (char *)&size, arg5_size);
	pack(bytes, bytes_off, lengths, lengths_off, (char *)&constant, arg6_size);
	pack(bytes, bytes_off, lengths, lengths_off, (char *)&global, arg7_size);

	//make the new syscall
	//syscall(273, 60, fatCubinHandle, hostVar, deviceAddress, deviceName, ext, size, constant, global);
	syscall(273, 60, num_args, lengths, bytes);

	delete bytes;
	delete lengths;
}


void __cudaRegisterShared(
                         void **fatCubinHandle,
                         void **devicePtr
                         )
{
	syscall(273, 61, fatCubinHandle, devicePtr);
}

void CUDARTAPI __cudaRegisterSharedVar(
                                      void   **fatCubinHandle,
                                      void   **devicePtr,
                                      size_t   size,
                                      size_t   alignment,
                                      int      storage
                                      )
{
	syscall(273, 62, fatCubinHandle, devicePtr, size, alignment, storage);
}

void __cudaRegisterTexture(
                          void **fatCubinHandle,
                          const struct textureReference *hostVar,
                          const void **deviceAddress,
                          const char *deviceName,
                          int dim,
                          int norm,
                          int ext
                          ) //passes in a newly created textureReference
{
	int num_args = 7;
	int arg0_size = sizeof(void **);
	int arg1_size = sizeof(char *);
	int arg2_size = sizeof(char *);
	int arg3_size = sizeof(const char *);
	int arg4_size = sizeof(int);
	int arg5_size = sizeof(int);
	int arg6_size = sizeof(int);

	int total_bytes = arg0_size + arg1_size + arg2_size + arg3_size + arg4_size +
							arg5_size + arg6_size;

	//allocate space for packed parameters and their lengths
	char *bytes = new char[total_bytes];
	int *lengths = new int[num_args];

	//variables to keep track of packed parameters and their lengths
	int bytes_off = 0;
	int lengths_off = 0;

	//pack up the parameters
	pack(bytes, bytes_off, lengths, lengths_off, (char *)&fatCubinHandle, arg0_size);
	pack(bytes, bytes_off, lengths, lengths_off, (char *)&hostVar, arg1_size);
	pack(bytes, bytes_off, lengths, lengths_off, (char *)&deviceAddress, arg2_size);
	pack(bytes, bytes_off, lengths, lengths_off, (char *)&deviceName, arg3_size);
	pack(bytes, bytes_off, lengths, lengths_off, (char *)&dim, arg4_size);
	pack(bytes, bytes_off, lengths, lengths_off, (char *)&norm, arg5_size);
	pack(bytes, bytes_off, lengths, lengths_off, (char *)&ext, arg6_size);

	//make the new syscall
	//syscall(273, 63, fatCubinHandle, hostVar, deviceAddress, deviceName, dim, norm, ext);
	syscall(273, 63, num_args, lengths, bytes);

	delete bytes;
	delete lengths;
}

#ifndef OPENGL_SUPPORT
typedef unsigned long GLuint;
#endif

cudaError_t cudaGLRegisterBufferObject(GLuint bufferObj)
{
	return (cudaError_t)syscall(273, 64, bufferObj);
}

cudaError_t cudaGLMapBufferObject(void** devPtr, GLuint bufferObj)
{
	return (cudaError_t)syscall(273, 65, devPtr, bufferObj);
}

cudaError_t cudaGLUnmapBufferObject(GLuint bufferObj)
{
	return (cudaError_t)syscall(273, 66, bufferObj);
}

cudaError_t cudaGLUnregisterBufferObject(GLuint bufferObj)
{
	return (cudaError_t)syscall(273, 67, bufferObj);
}

#if (CUDART_VERSION >= 2010)

cudaError_t CUDARTAPI cudaHostAlloc(void **pHost,  size_t bytes, unsigned int flags)
{
	return (cudaError_t)syscall(273, 68, pHost, bytes, flags);
}

cudaError_t CUDARTAPI cudaHostGetDevicePointer(void **pDevice, void *pHost, unsigned int flags)
{
	return (cudaError_t)syscall(273, 69, pDevice, pDevice, flags);
}

cudaError_t CUDARTAPI cudaSetValidDevices(int *device_arr, int len)
{
	return (cudaError_t)syscall(273, 70, device_arr, len);
}

cudaError_t CUDARTAPI cudaSetDeviceFlags( int flags )
{
	return (cudaError_t)syscall(273, 71, flags);
}

cudaError_t CUDARTAPI cudaFuncGetAttributes(struct cudaFuncAttributes *attr, const char *func)
{
	return (cudaError_t)syscall(273, 72, attr, func);
}

cudaError_t CUDARTAPI cudaEventCreateWithFlags(cudaEvent_t *event, int flags)
{
	return (cudaError_t)syscall(273, 73, event, flags);
}

cudaError_t CUDARTAPI cudaDriverGetVersion(int *driverVersion)
{
	return (cudaError_t)syscall(273, 74, driverVersion);
}

cudaError_t CUDARTAPI cudaRuntimeGetVersion(int *runtimeVersion)
{
	return (cudaError_t)syscall(273, 75, runtimeVersion);
}

#endif

cudaError_t CUDARTAPI cudaGLSetGLDevice(int device)
{
	return (cudaError_t)syscall(273, 76, device);
}

typedef void* HGPUNV;

cudaError_t CUDARTAPI cudaWGLGetDevice(int *device, HGPUNV hGpu)
{
	return (cudaError_t)syscall(273, 77, device, hGpu);
}

void CUDARTAPI __cudaMutexOperation(int lock)
{
	syscall(273, 78, lock);
}

void  CUDARTAPI __cudaTextureFetch(const void *tex, void *index, int integer, void *val)
{
	syscall(273, 79, tex, index, integer, val);
}

}

namespace cuda_math {

void CUDARTAPI __cudaMutexOperation(int lock)
{
	syscall(273, 78, lock);
}

void  CUDARTAPI __cudaTextureFetch(const void *tex, void *index, int integer, void *val)
{
	syscall(273, 79, tex, index, integer, val);
}

int CUDARTAPI __cudaSynchronizeThreads(void** a, void* b)
{
	return (int)syscall(273, 80, a, b);
}

}

